#include <linux/syscalls.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/kallsyms.h>
#include <linux/cdev.h>
#include <linux/version.h>
#include <linux/ptrace.h>		// force_successful_syscall_return defined here in 2.6.33
#include <linux/device.h>		// class_create defined here in 2.6.33

#include <asm/tlbflush.h>

#include "pgmod.h"

static struct file_operations fops;

MODULE_AUTHOR("Henry Wong <henry@stuffedcow.net>");
MODULE_DESCRIPTION("Create /proc/pgmod to allow retrieving CR3 and CR4 and flushing TLB. Create /dev/pgmod_mem to allow unlimited physical memory access.");
MODULE_LICENSE("GPL");



static int pgmod_mem_init(void);
static void pgmod_mem_cleanup(void);

#ifndef LINUX_VERSION_CODE
#error "Why is LINUX_VERSION_CODE not defined?"
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
#define my_on_each_cpu(func,info,wait) on_each_cpu(func,info,wait)
#else
#define my_on_each_cpu(func,info,wait) on_each_cpu(func,info,0,wait)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25)
#error "Your kernel is too old. I don't know how to replace proc_create(), which first appeared in kernel 2.6.25."
#endif

static inline uintptr_t get_cr3(void)
{
	uintptr_t val;
	asm volatile("mov %%cr3,%0\n" : "=r"(val), "=m"(__force_order));
	return val;
}
static inline uintptr_t get_cr4(void)
{
	uintptr_t val;
	asm volatile("mov %%cr4,%0\n" : "=r"(val), "=m"(__force_order));
	return val;
}
/*static inline uint32_t cpuid_edx(uint32_t leaf) 
{
	uint32_t edx;
	asm volatile("cpuid" : "=d"(edx) : "a"(leaf) : "ecx","ebx");
	return edx;
}*/

static inline uint64_t my_rdmsr(unsigned int msr)
{
	uint32_t hi,lo;
	asm volatile("rdmsr" : "=a"(lo), "=d"(hi): [msr] "c"(msr));
	return ((uint64_t)hi<<32ULL) | lo;
}

static void do_flush_tlb_all(void *info)
{
	// Reload cr3 without modification, to flush TLB (except Global entries)
	unsigned long val;
	asm volatile("mov %%cr3, %0\n"
		"mov %0, %%cr3" : "=r"(val), "+m"(__force_order) );
}
static void do_invlpg(void *addr)
{
	asm volatile("invlpg (%0)\n" : : "r"(addr));
}
static void do_wbinvd(void *info)
{
	asm volatile ("wbinvd");
}

//static unsigned char old_code[8];
//static int(*p_devmem_is_allowed)(unsigned long) = 0;
//static int (*p_set_memory_rw)(unsigned long addr, int numpages) = 0;

static int __init pgmod_init(void)
{
	printk(KERN_INFO "Starting pgmod\n");
	
	proc_create("pgmod", 0666,NULL,&fops);
	pgmod_mem_init();
	return 0;
}

static void __exit pgmod_cleanup(void)
{
	pgmod_mem_cleanup();
	remove_proc_entry("pgmod", NULL);
	printk(KERN_INFO "Unloaded pgmod\n");
}


static int pgmod_proc_open(struct inode *sp_inode, struct file *sp_file)
{
	//printk(KERN_INFO "proc_open");
	return 0;
}
static int pgmod_proc_release(struct inode *sp_inode, struct file *sp_file)
{
	//printk(KERN_INFO "proc_release");
	return 0;
}
static ssize_t pgmod_proc_read(struct file *sp_file, char __user *buf, size_t size, loff_t *offset)
{
	char kbuf[64];		// Make sure the string doesn't overflow this.
	int len=0;
	
	if (*offset>=1)
	{
		return 0;
	}

	len += sprintf(kbuf+len, "%llx\n", (uint64_t)get_cr3());
	len += sprintf(kbuf+len, "%llx\n", (uint64_t)get_cr4());
	if (cpuid_edx(1) & 0x10000)
		len += sprintf(kbuf+len, "%llx\n", my_rdmsr(0x277));
	else
		len += sprintf(kbuf+len, "%llx\n", 0x0007040600070406ULL);
	len++;	// NULL-termination
		
	if (copy_to_user(buf, kbuf, len))
		printk(KERN_INFO "copy_to_user failed");

	(*offset)++;		// Advance offset by 1 arbitrary unit.

	return len;
}
static long pgmod_proc_ioctl(struct file *sp_file, unsigned int ioctl_num, unsigned long ioctl_param)
{
	long retval = 0;
	switch (ioctl_num) {
		case PGMOD_IOCTL_WBINVD:
			if (ioctl_param > 0)
				my_on_each_cpu(do_wbinvd, NULL, 1);
			else
				do_wbinvd(NULL);
			break;
		case PGMOD_IOCTL_FLUSHTLB:
			if (ioctl_param > 0)
				my_on_each_cpu(do_flush_tlb_all, NULL, 1);
			else
				do_flush_tlb_all(NULL);
			break;
		case PGMOD_IOCTL_INVLPG:
			do_invlpg((void*)ioctl_param);
			break;
		case PGMOD_IOCTL_NULL:
			break;
		case PGMOD_IOCTL_LOAD:
			{
				unsigned int lv;
				asm volatile ("movl (%[addr]), %[retval]\n" : [retval] "=r"(lv) : [addr] "r"(ioctl_param));
				retval = lv;
			}
			break;
		default:
			printk(KERN_INFO "pgmod: Unknown ioctl\n");
	}
	return retval;
}

static ssize_t pgmod_proc_write(struct file *sp_file, const char __user *buf, size_t size, loff_t *offset)
{
	//printk(KERN_INFO "proc_write: Flushing all TLBs");
	my_on_each_cpu(do_flush_tlb_all, NULL, 1);

/*
	if (!p_devmem_is_allowed) {
		p_devmem_is_allowed = (int(*)(unsigned long)) kallsyms_lookup_name("devmem_is_allowed");
		if (p_devmem_is_allowed) {
			printk(KERN_INFO "Patching devmem_is_allowed() at 0x%p\n", p_devmem_is_allowed);
		
			p_set_memory_rw = (int(*)(unsigned long, int)) kallsyms_lookup_name("set_memory_rw");
			printk(KERN_INFO "set_memory_rw is at 0x%p\n", p_set_memory_rw);
			if (p_set_memory_rw) {
				int rv=1234;				
				int (*p_printk)(const char*);
				int (*p_test)(void);

				p_printk = (int (*)(const char*)) kallsyms_lookup_name("printk");
				if (p_printk) {
					printk(KERN_INFO "printk is at %p\n", p_printk);
					p_printk(KERN_INFO "Testing the printk pointer\n");
				}

				p_test = (int (*)(void)) kallsyms_lookup_name("pgmod_test_func");
				if (p_test) {
					printk(KERN_INFO "test_func is at %p\n", p_test);
					p_test();
				}

				
				printk(KERN_INFO "Enabling read-write permissions on devmem_is_allowed\n");
				rv = p_set_memory_rw((unsigned long)p_devmem_is_allowed & (~0xfffUL), 1);
				printk(KERN_INFO "set_memory_rw returned %d\n", rv);
				memcpy (old_code, p_devmem_is_allowed, sizeof(old_code));
				
				// Why does kallsyms_lookup_name return a different pointer than 
				// /proc/kallsyms? Calling the function still works somehow, but
				// set_memory_rw doesn't. Trying to modify the code at the pointer
				// results in a fault at the "real" address (the one returned by /proc/kallsyms)
				
				//printk(KERN_INFO "Attempting to write to code at 0x%p\n", p_devmem_is_allowed);
				//memcpy (p_devmem_is_allowed, old_code, sizeof(old_code));
			}
			else {
				printk(KERN_INFO "Couldn't find symbol set_memory_rw.\n");
				p_devmem_is_allowed = 0;
			}
		}	
	}
*/

	return size;
}


static struct file_operations fops = {
	.open = pgmod_proc_open,
	.read = pgmod_proc_read,
	.write = pgmod_proc_write,
	.release = pgmod_proc_release,
	.unlocked_ioctl = pgmod_proc_ioctl
};

module_init(pgmod_init);
module_exit(pgmod_cleanup);











// Simplified from linux/drivers/char/mem.c
#define PGMOD_MEM_DEVICE_NAME "pgmod_mem"
static dev_t pgmod_mem_devnum = 0;
static struct class *pgmod_mem_class=0;
static struct device *pgmod_mem_dev=0;
static struct cdev *pgmod_mem_cdev=0;

static void* (*p_xlate_dev_mem_ptr)(phys_addr_t phys) = 0;
static void (*p_unxlate_dev_mem_ptr)(phys_addr_t phys, void *addr) = 0;

static inline unsigned long size_inside_page(unsigned long start,
					     unsigned long size)
{
	unsigned long sz;

	sz = PAGE_SIZE - (start & (PAGE_SIZE - 1));

	return min(sz, size);
}

static int pgmod_mem_open(struct inode *inode, struct file *filp)
{
	return capable(CAP_SYS_RAWIO) ? 0 : -EPERM;
}

/*
 * The memory devices use the full 32/64 bits of the offset, and so we cannot
 * check against negative addresses: they are ok. The return value is weird,
 * though, in that case (0).
 *
 * also note that seeking relative to the "end of file" isn't supported:
 * it has no meaning, so it returns -EINVAL.
 */
static loff_t pgmod_mem_lseek(struct file *file, loff_t offset, int orig)
{
	loff_t ret;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,5,0)
	inode_lock(file_inode(file));
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0)
	mutex_lock(&file_inode(file)->i_mutex);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	mutex_lock(&file->f_path.dentry->d_inode->i_mutex);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
	mutex_lock(&file->f_dentry->d_inode->i_mutex);
#else	// Before 2.6.16
	down(&file->f_dentry->d_inode->i_sem);
#endif
	switch (orig) {
	case SEEK_CUR:
		offset += file->f_pos;
	case SEEK_SET:
		/* to avoid userland mistaking f_pos=-9 as -EBADF=-9 */
		if ((unsigned long long)offset >= -MAX_ERRNO) {
			ret = -EOVERFLOW;
			break;
		}
		file->f_pos = offset;
		ret = file->f_pos;
		force_successful_syscall_return();
		break;
	default:
		ret = -EINVAL;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,5,0)
	inode_unlock(file_inode(file));		// 4.15, 4.17, 4.10, 4.6, 4.5
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0)
	mutex_unlock(&file_inode(file)->i_mutex);		// 4.0.0, 3.19, 3.9, 4.3, 4.4, 4.4.131
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20)
	mutex_unlock(&file->f_path.dentry->d_inode->i_mutex);	// 3.6.6, 3.0, 3.3, 3.7, 3.8, 2.6.39, 2.6.24, 2.6.20
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
	mutex_unlock(&file->f_dentry->d_inode->i_mutex);		//	 2.6.16, 2.6.18, 2.6.19
#else
	up(&file->f_dentry->d_inode->i_sem);		// 2.6.11, 2.6.14, 2.6.15
#endif
	return ret;
}


/*
 * This funcion reads the *physical* memory. The f_pos points directly to the
 * memory location.
 */
static ssize_t pgmod_mem_read(struct file *file, char __user *buf,
			size_t count, loff_t *ppos)
{
	phys_addr_t p = *ppos;
	ssize_t read, sz;
	void *ptr;

	if (p != *ppos)
		return 0;
		
	if (!p_xlate_dev_mem_ptr || !p_unxlate_dev_mem_ptr)
		return 0;

	read = 0;
#ifdef __ARCH_HAS_NO_PAGE_ZERO_MAPPED
	/* we don't have page 0 mapped on sparc and m68k.. */
	if (p < PAGE_SIZE) {
		sz = size_inside_page(p, count);
		if (sz > 0) {
			if (clear_user(buf, sz))
				return -EFAULT;
			buf += sz;
			p += sz;
			count -= sz;
			read += sz;
		}
	}
#endif

	while (count > 0) {
		unsigned long remaining;
		int allowed;

		sz = size_inside_page(p, count);

		allowed = 1;
		if (!allowed)
			return -EPERM;
		if (allowed == 2) {
			/* Show zeros for restricted memory. */
			remaining = clear_user(buf, sz);
		} else {
			/*
			 * On ia64 if a page has been mapped somewhere as
			 * uncached, then it must also be accessed uncached
			 * by the kernel or data corruption may occur.
			 */
			ptr = p_xlate_dev_mem_ptr(p);
			if (!ptr)
				return -EFAULT;

			remaining = copy_to_user(buf, ptr, sz);

			p_unxlate_dev_mem_ptr(p, ptr);
		}

		if (remaining)
			return -EFAULT;

		buf += sz;
		p += sz;
		count -= sz;
		read += sz;
	}

	*ppos += read;
	return read;
}

static ssize_t pgmod_mem_write(struct file *file, const char __user *buf,
			 size_t count, loff_t *ppos)
{
	phys_addr_t p = *ppos;
	ssize_t written, sz;
	unsigned long copied;
	void *ptr;

	if (p != *ppos)
		return -EFBIG;

	if (!p_xlate_dev_mem_ptr || !p_unxlate_dev_mem_ptr)
		return 0;

	written = 0;

#ifdef __ARCH_HAS_NO_PAGE_ZERO_MAPPED
	/* we don't have page 0 mapped on sparc and m68k.. */
	if (p < PAGE_SIZE) {
		sz = size_inside_page(p, count);
		/* Hmm. Do something? */
		buf += sz;
		p += sz;
		count -= sz;
		written += sz;
	}
#endif

	while (count > 0) {
		int allowed;

		sz = size_inside_page(p, count);

		allowed = 1;
		if (!allowed)
			return -EPERM;

		/* Skip actual writing when a page is marked as restricted. */
		if (allowed == 1) {
			/*
			 * On ia64 if a page has been mapped somewhere as
			 * uncached, then it must also be accessed uncached
			 * by the kernel or data corruption may occur.
			 */
			ptr = p_xlate_dev_mem_ptr(p);
			if (!ptr) {
				if (written)
					break;
				return -EFAULT;
			}

			copied = copy_from_user(ptr, buf, sz);
			p_unxlate_dev_mem_ptr(p, ptr);
			if (copied) {
				written += sz - copied;
				if (written)
					break;
				return -EFAULT;
			}
		}

		buf += sz;
		p += sz;
		count -= sz;
		written += sz;
	}

	*ppos += written;
	return written;
}

static struct file_operations mem_fops = {
	.llseek		= pgmod_mem_lseek,
	.read		= pgmod_mem_read,
	.write		= pgmod_mem_write,
//	.mmap		= mmap_mem,
	.open		= pgmod_mem_open
};


static char *pgmod_mem_devnode(struct device *dev, umode_t *mode)
{
	if (mode)
		*mode = 0;
	return NULL;
}

static int pgmod_mem_init(void)
{
	//printk(KERN_INFO "Initializing /dev/pgmod_mem device\n");
	
	if (alloc_chrdev_region(&pgmod_mem_devnum, 0, 1, PGMOD_MEM_DEVICE_NAME) < 0) {
		printk (KERN_INFO "alloc_chrdev_region failed\n");
		pgmod_mem_devnum = 0;
		return -1;
	}
	
	pgmod_mem_class = class_create(THIS_MODULE, PGMOD_MEM_DEVICE_NAME);
	if (IS_ERR(pgmod_mem_class)) {
		printk(KERN_INFO "Couldn't register device class\n");
		return PTR_ERR(pgmod_mem_class);
	}

	pgmod_mem_cdev = cdev_alloc();
	if (!pgmod_mem_cdev) {
		printk(KERN_INFO "cdev_alloc failed\n");
		return -1;
	}
	cdev_init(pgmod_mem_cdev, &mem_fops);

	if (cdev_add(pgmod_mem_cdev, pgmod_mem_devnum, 1) < 0) {
		printk(KERN_INFO "cdev_add failed\n");
		return -1;
	}

	pgmod_mem_class->devnode = pgmod_mem_devnode;
	pgmod_mem_dev = device_create(pgmod_mem_class, NULL, pgmod_mem_devnum,
			      NULL, PGMOD_MEM_DEVICE_NAME);
     if (IS_ERR(pgmod_mem_dev)) {
		printk(KERN_INFO "Couldn't create device\n");
		return PTR_ERR(pgmod_mem_dev);
	}

	p_xlate_dev_mem_ptr = (void*(*)(phys_addr_t)) kallsyms_lookup_name("xlate_dev_mem_ptr");
	p_unxlate_dev_mem_ptr = (void(*)(phys_addr_t,void*)) kallsyms_lookup_name("unxlate_dev_mem_ptr");
	printk(KERN_INFO "Function xlate_dev_mem_ptr is at %p\n", p_xlate_dev_mem_ptr);
	printk(KERN_INFO "Function unxlate_dev_mem_ptr is at %p\n", p_unxlate_dev_mem_ptr);

	printk(KERN_INFO "Created /dev/pgmod_mem with major number %d\n", MAJOR(pgmod_mem_devnum));

	return 0;	
}
static void pgmod_mem_cleanup(void)
{
	if (!IS_ERR(pgmod_mem_dev) && !IS_ERR(pgmod_mem_class)) {
	//	printk(KERN_INFO "Calling device_destroy\n");
		device_destroy(pgmod_mem_class, pgmod_mem_devnum);
	}
	
	if (pgmod_mem_cdev) {
		cdev_del(pgmod_mem_cdev);
	}

	if (!IS_ERR(pgmod_mem_class)) {
	//	printk(KERN_INFO "Calling class_destroy\n");	
		class_destroy(pgmod_mem_class);
	}

	if (pgmod_mem_devnum) {
	//	printk(KERN_INFO "Calling unregister_chrdev_region\n");
		unregister_chrdev_region(pgmod_mem_devnum,1);
	}
	//printk(KERN_INFO "Cleaning up /dev/pgmod_mem device\n");	
}











