#ifndef PGUTIL_H
#define PGUTIL_H

// Via Nano behaves differently depending on whether the result of a load goes to a load or to an ALU uop.
// VIA_NANO=1 inserts a add $0 so that the load result goes through an ALU add before going back to the next load.
#ifndef VIA_NANO
#define VIA_NANO 0
#endif

#ifndef OLD_CPU
#define OLD_CPU 0		// Assume no clflush, mfence, prefetch, ...
#endif

#ifndef PGMODE
#define PGMODE 64
#endif



#if ((PGMODE==64) || (PGMODE==36))
	#define _LARGEFILE64_SOURCE
	#define _FILE_OFFSET_BITS 64
#else 
	#define lseek64 lseek	
#endif

#include <stdint.h>


#if (PGMODE==64)
	typedef uint64_t pte_t;	// 64-bit always uses 8-byte PTEs
	typedef uint64_t reg_t;	// 64-bit integer registers
	typedef uint64_t paddr_t;	// 52-bit physical address
	int pte_per_pde = 512;		// 512 for PAE/64bit, 1024 for 32-bit.
	const int translation_levels = 4;
	const int translation_shifts[4] = { 12, 21, 30, 39 };
	#warning "Using 64-bit paging"
#elif (PGMODE==36)
	typedef uint32_t reg_t;	// 32-bit integer registers
	typedef uint64_t pte_t;	// 32-bit PAE uses 8-byte PTEs
	typedef uint64_t paddr_t;	// 36-bit physical address
	int pte_per_pde = 512;		// 512 for PAE/64bit, 1024 for 32-bit.
	const int translation_levels = 3;
	const int translation_shifts[3] = { 12, 21, 30 };
	#warning "Using 32-bit PAE paging"
#elif (PGMODE==32)
	typedef uint32_t reg_t;	// 32-bit integer registers
	typedef uint32_t pte_t;	// 32-bit non-PAE uses 4-byte PTEs.
	typedef uint32_t paddr_t;	// 32-bit physical address
	int pte_per_pde = 1024;		// 512 for PAE/64bit, 1024 for 32-bit.
	const int translation_levels = 2;
	const int translation_shifts[2] = { 12, 22 };
	#warning "Using 32-bit paging"
#else
	#error "Unknown architecture. Set PGMODE to 64, 36 (PAE), or 32."	
#endif



typedef struct {
	paddr_t cr3;
	reg_t cr4;
	uint64_t patmsr;
	int pat;
} pgutil_info_struct;





#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/mman.h>

#define __USE_GNU
#include <signal.h>
#include <ucontext.h>

#include "pgmod.h"
#include <sys/ioctl.h>


static int fd_pgmod=-1;
static int fd_devmem = -1;
static int allow_1gb_pages = 0;


inline uint64_t rdtscp()
{
	uint32_t lo, hi;

	__asm__ volatile (".byte 0x0f, 0x01, 0xf9" : "=a" (lo), "=d" (hi) : : "ecx");
	return (uint64_t)(((uint64_t)hi)<<32LL) | (uint64_t) lo;
}

inline uint64_t rdtscp32()
{
	uint32_t lo;

	__asm__ volatile (".byte 0x0f, 0x01, 0xf9" : "=a" (lo) : : "edx","ecx");
	return lo;
}

inline uint64_t rdtsc()
{
	uint32_t lo, hi;

	__asm__ volatile (".byte 0x0f, 0x31" : "=a" (lo), "=d" (hi));
	return (uint64_t)(((uint64_t)hi)<<32LL) | (uint64_t) lo;
}

inline uint64_t rdtsc32()
{
	uint32_t lo;

	__asm__ volatile (".byte 0x0f, 0x31" : "=a" (lo) : : "edx");
	return lo;
}



void read_physical(paddr_t addr, void* buf, size_t size)
{
	if (fd_devmem < 0)
		return;
		
	lseek64(fd_devmem, addr, SEEK_SET);
	if (size != read(fd_devmem, buf, size)) {
		printf ("read_physical from 0x%016llx failed\n", (uint64_t)addr);
	}
}

void write_physical(paddr_t addr, const void* buf, size_t size)
{
	if (fd_devmem < 0)
		return;
	if (addr < 65536)
		printf ("Something seems wrong. Writing to address 0x%llu refused.\n", (uint64_t)addr);
		
	lseek64(fd_devmem, addr, SEEK_SET);
	write(fd_devmem, buf, size);
}


void pgmod_flush_tlb()
{
	ioctl (fd_pgmod, PGMOD_IOCTL_FLUSHTLB, 1);		// Flush TLB on all CPUs
}
void pgmod_wbinvd()
{
	ioctl (fd_pgmod, PGMOD_IOCTL_WBINVD, 0);		// Flush cache on this CPU only
}
void pgmod_invlpg(void* addr)
{
	ioctl (fd_pgmod, PGMOD_IOCTL_INVLPG, (unsigned long)addr);		// Invalidate page on current CPU.
}
void pgmod_null_ioctl()
{
	ioctl (fd_pgmod, PGMOD_IOCTL_NULL, 0);
}
long pgmod_kernel_load(void *address)
{
	return ioctl (fd_pgmod, PGMOD_IOCTL_LOAD, (unsigned long)address);
}
void pgmod_ioctl(unsigned int request, unsigned long parameter)
{
	ioctl (fd_pgmod, request, parameter);		// Flush cache on this CPU only
}
void excessive_fence()
{
#if OLD_CPU
	for (int i=0;i<8;i++) {
		asm volatile ( R"(
			call return_2_train
			call return_2_train		# Train the RAS (for Via Nano)
			call return_2; 1: jmp 1b		# Pipeline flush
		)" );
	}
#else
	for (int i=0;i<4;i++) {
		asm volatile (
			"mfence\n"
			"mfence\n"
			"mfence\n"
			"mfence\n"
			"mfence\n"
			"mfence\n"
			"mfence\n"
			"mfence\n"
		);
	}
#endif
}



// Translates virtual address to physical.
// 4KB paging in 32-bit, PAE, and 64-bit modes are similar enough this can be done with 
// one routine and a bunch of architecture-specific typedefs.
// If level > 0, stop early by "level" lookups and return the PTE. Useful for getting
//   a phyaddr pointer to a page table to edit it.
//   0 : Do a complete translation
//   1 : Stop at the final PTE. Return the final PTE
//   2 : Return the PDE, one level up the tree
// If level == 0, return the completed translated address, including page offset.
// If level < 0, return the physical address of the PTE/PDE/PML3 at -level. 
//   -1 : Return physical address of the PTE entry
//   -2 : Return physical address of PDE entry
// Large pages: level is relative to the page table walk, so level = 1 is always
// the final PTE/PDE/PMl3 and depends on the page size.
pte_t translate(paddr_t cr3, uintptr_t vaddr, int level)
{
	// Canonical form only applies to 64-bit pointers with 48-bit virtual address space.
	if ((sizeof(vaddr) == 8) && (uint64_t)vaddr >> 48 != 0 && (uint64_t)vaddr >> 48 != 0xffff)
	{
		printf ("vaddr %016llx is not in canonical form\n", (uint64_t)vaddr);
		return 0;
	}

	int l;
	pte_t pte[4];
	paddr_t pte_addr[4];
	paddr_t cur_ptr = cr3;
	
	for (l = translation_levels-1; l >= 0; l--)
	{
		int cur_offset = (((paddr_t)vaddr) >> translation_shifts[l]) & (pte_per_pde-1);
		pte_addr[l] = cur_ptr + cur_offset*sizeof(pte_t);
		read_physical(pte_addr[l], &pte[l], sizeof(pte_t));
		
		//printf ("  Read: %016llx at level %d from %016llx\n", (uint64_t)pte[l], l, (uint64_t)pte_addr[l]);

		if ((l == 1 || (allow_1gb_pages && l==2)) && (pte[l] & 0x80)) {
			//printf ("Large page at level %d detected.\n", l);
			break;
		}

		if (~pte[l] & 1) {
			printf ("Page table level %d entry %016llx not present. Fail to translate %016llx\n", l, (uint64_t)pte[l], (uint64_t)vaddr);
			exit(1);
			return 0;
		}
		cur_ptr = pte[l] & 0x000ffffffffff000ULL;
	}
	
	if (l < 0) l = 0;
	
	//printf ("Translation finished at level %d\n", l);
	if (level > 0) {
		return pte[l+level-1];
	} else if (level < 0) {
		return pte_addr[l-1-level];		// Return physical address of the PTE at a particular level
	} else {
		return (pte[l] & (0x0010000000000000 - (1ULL<<translation_shifts[l]))) | (vaddr & ((1ULL<<translation_shifts[l])-1));
	}
}

// Changes a PTE. Returns the old PTE. Only for the lowest level of page tables.
pte_t change_pte(paddr_t cr3, uintptr_t vaddr, pte_t new_pte)
{
	paddr_t p_pte = translate(cr3, vaddr, -1);
		
	pte_t old_pte; 
	read_physical(p_pte, &old_pte, sizeof(pte_t));
	write_physical(p_pte, &new_pte, sizeof(pte_t));
	printf ("Updating PTE from %016llx to %016llx, at %016llx\n", (uint64_t)old_pte, (uint64_t)new_pte, (uint64_t)p_pte);
	
	return old_pte;
}

int pgutil_init(pgutil_info_struct *pginfo)
{
	/* Use pgmod to set up the test
	   /proc/pgmod - Used to retrieve CR3 and CR4, and reload CR3 to cause a TLB flush
	   /dev/pgmod_mem - Same as /dev/mem but bypasses all security restrictions (e.g., CONFIG_STRICT_DEVMEM)
	   
	   This is used to allow control over the page tables. We allocate two memory pages (using mmap), then
	   change the page tables for that array so both virtual pages point to the same physical page. This is the
	   "victim" page that contains data we want to read. The first virtual page's permissions are changed to see
	   how permissions affects the ability to read the victim. The second virtual page's permissions are
	   not restricted, and is used to preload/pin the physical page into various (mainly L1) caches.
	   
	   On Intel CPUs, the ability to read from the victim page where permissions checks fail requires that the
	   data actually be in the L1 cache. The second page is used to ensure this is the case.
	*/
	
	FILE *f = fopen("/proc/pgmod","rt");
	if (!f) {
		perror("Can't open /proc/pgmod. Module not loaded?");
		return -1;
	}
	
	fd_devmem = open("/dev/pgmod_mem", O_RDWR | O_SYNC);
	if (fd_devmem == -1)
	{
		perror ("Can't open /dev/pgmod_mem. Permissions?");
		return -2;
	}
		
	uint64_t cr4, patmsr, tmp;
	paddr_t cr3;	
	fscanf(f, "%llx ", &tmp); cr3 = (paddr_t)tmp;
	fscanf(f, "%llx ", &cr4);
	fscanf(f, "%llx ", &patmsr);	
	fclose(f);

	fd_pgmod = open("/proc/pgmod", O_NONBLOCK|O_RDONLY);
	if (fd_pgmod == -1) {
		perror("Opening /proc/pgmod");
		return -1;
	}
	
	

	unsigned int cpuid_edx=0, cpuid_80000001_edx=0;
	// Detect PAT, 1GB pages
	{
		asm volatile ("mov $1, %%eax; cpuid\n" : "=d"(cpuid_edx) : : "eax","ebx","ecx");
#if PGMODE==64
		asm volatile ("mov $0x80000001, %%eax; cpuid\n" : "=d"(cpuid_80000001_edx) : : "eax","ebx","ecx");		
#endif		
		printf ("CPUID:\n");
		printf ("  PAT %s\n", (cpuid_edx & 0x10000) ? "enabled" : "not supported");
		if (~cpuid_edx & 0x10000) {
			patmsr = 0x0007040600070406;
		}
		printf ("  1 GB pages %s\n", (cpuid_80000001_edx & 0x4000000) ? "enabled" : "not supported");
		if (cpuid_80000001_edx & 0x4000000) allow_1gb_pages = 1;
	}

	printf ("Compiled for %s usermode with %s paging\n", 
#if __i386__
	"32-bit",
#elif __x86_64__
	"64-bit",
#endif	
		translation_levels == 4 ? "x86-64" : translation_levels == 3 ? "x86 PAE" : "x86 non-PAE");
	printf ("CR3 appears to be %016llx\n", (uint64_t)cr3);
	printf ("  translation levels = %d\n", translation_levels);
	printf ("CR4 appears to be %016llx\n", (uint64_t)cr4);
	printf ("  PAE = %s\n", (cr4 & 0x20) ? "Enabled" : "Disabled");
	printf ("  LA57 = %s\n", (cr4 & 0x1000) ? "Enabled" : "Disabled");
	printf ("  PCIDE = %s\n", (cr4 & 0x20000) ? "Enabled" : "Disabled");		// Context ID support
	printf ("PAT MSR = %016llx\n", (uint64_t)patmsr);
	if (cr4 & 0x1000) {
		printf ("LA57 5-level paging not supported. Quitting.\n");
		return -3;
	}
	if ((pte_per_pde == 512) != !!(cr4 & 0x20)) {
		printf ("PAE state unexpected. Quitting.\n");		// PAE has 512 PTEs per page table, non-PAE has 1024.
		return -3;
	}
	if (cr4 & 0x20000) {
		cr3 &= ~0xfffULL;		// Remove the process context ID
	}
	
	if (pginfo) {
		pginfo->cr3 = cr3;
		pginfo->cr4 = (reg_t)cr4;
		pginfo->patmsr = patmsr;
		pginfo->pat = !!(cpuid_edx&0x10000);
	}
	
	return 0;
}

void pgutil_cleanup()
{
	close(fd_devmem);
	close(fd_pgmod);
	fd_devmem = -1;
	fd_pgmod = -1;
}


#endif
