/* 
  Henry Wong <henry@stuffedcow.net>
  http://blog.stuffedcow.net/2018/05/meltdown-microarchitecture/
  
  Test for speculative load when reading from page with various permissions.
  Requires pgmod.
  
  2018-05-07
  
  This program contains both 32-bit and 64-bit x86 inline assembly, which I hope are equivalent.
  Must manually set which paging mode (64, PAE32, 32) is in use by #defining PGMODE (see pgutil.h).
  A 32-bit user-mode program can be running on an OS using 32-bit, 32-bit PAE, or 64-bit paging.
  It is possible to detect the paging mode at run time, but I was lazy.
*/

#include "pgutil.h"

#define ITS 32768

#define SPECLOAD_STATUS 1
#define SPECLOAD_VALUE 2
#define SPECLOAD_REPORT SPECLOAD_STATUS	// SPECLOAD_STATUS or SPECLOAD_VALUE. SPECLOAD_VALUE may fail if hardware prefetching is too aggressive.

#define ALIAS_DISTANCE (1048576)		// In units of 4 bytes. Must be multiple of 1024 (page size)
			// Pentium 4 Prescott and newer wants a multiple of 4MB (related to 4M aliasing: Using 22-bit virtual address tags in the L1)
			// Pentium 4 Willamette/Northwood has 64K aliasing, so presumably needs to be a multiple of 64KB, but I haven't tested this.
#define DONT_REMAP_PTE1 0
#define DONT_TRUST_CLFLUSH 1		// At least Pentium 4 Presler: clflush doesn't flush (possibly won't flush different virtual address?)


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/mman.h>

#include <math.h>		// For sqrt
#include <setjmp.h>		// For cleanup.
#include <sched.h>		// For sched_yield.



#define STR_(x) #x
#define STR(x) STR_(x)
#define STR4(x) STR(4*(x))

#define REPH(N) REP_##N
#define REP(N,X) REPH(N)(X)
#define REP_0(X)
#define REP_1(X) X
#define REP_2(X) X X
#define REP_3(X) X X X
#define REP_4(X) X X X X
#define REP_5(X) X X X X X
#define REP_6(X) X X X X X X
#define REP_7(X) X X X X X X X
#define REP_8(X) X X X X X X X X
#define REP_9(X) REP_4(X) REP_5(X)
#define REP_10(X) REP_5(X) REP_5(X)
#define REP_11(X) REP_5(X) REP_6(X)
#define REP_12(X) REP_6(X) REP_6(X)
#define REP_13(X) REP_6(X) REP_7(X)
#define REP_14(X) REP_7(X) REP_7(X)
#define REP_15(X) REP_7(X) REP_8(X)
#define REP_16(X) REP_8(X) REP_8(X)
#define REP_17(X) REP_8(X) REP_9(X)
#define REP_18(X) REP_9(X) REP_9(X)
#define REP_19(X) REP_9(X) REP_10(X)
#define REP_20(X) REP_10(X) REP_10(X)
#define REP_21(X) REP_10(X) REP_11(X)
#define REP_22(X) REP_11(X) REP_11(X)
#define REP_23(X) REP_11(X) REP_12(X)
#define REP_24(X) REP_12(X) REP_12(X)
#define REP_25(X) REP_12(X) REP_13(X)
#define REP_26(X) REP_13(X) REP_13(X)
#define REP_27(X) REP_13(X) REP_14(X)
#define REP_28(X) REP_14(X) REP_14(X)
#define REP_29(X) REP_14(X) REP_15(X)
#define REP_30(X) REP_15(X) REP_15(X)
#define REP_31(X) REP_15(X) REP_16(X)
#define REP_32(X) REP_16(X) REP_16(X)
#define REP_36(X) REP_18(X) REP_18(X)
#define REP_48(X) REP_24(X) REP_24(X)
#define REP_64(X) REP_32(X) REP_32(X)
#define REPH2(X) REPH(X)

#define DELAYAMT 4		// Caution: Core 2 needs a high DELAYAMT of at least ~26. Ivy Bridge wants low DELAYAMT (4). In-order memory wants a different delay method.
#define DELAY_OFFSET 768	// Choose this to avoid colliding in the same set with victim or target array
void** delay_array;
void* delay_array_pointer;

#if __i386__
asm(R"(
	.text
	.align 4096; 
			
	delay_and_return:
		add %eax, (%esp)
)"
		"mov delay_array_pointer, %eax\n"
		REPH2(DELAYAMT) ("mov (%eax), %eax\n")
		"mov %eax, delay_array_pointer\n"	
#if 0
	REPH2(DELAYAMT) (
		"clflush delay_array_pointer\n"
		"addl $0, delay_array_pointer\n"
		"clflush delay_array_pointer\n"
		"addl $0, delay_array_pointer+1\n" )		
#endif	
R"(		
		and $0x1, %eax
		add %eax, (%esp)
	delay_and_return_train:
		ret
		
		return_2:
			addl $2, (%esp)
		return_2_train:	
			ret
)");
#elif __x86_64__
asm(R"(
	.text
	.align 4096; delay_and_return:
		add %rax, (%rsp)
)"
#if 1
		"mov delay_array_pointer, %rax\n"
		REPH2(DELAYAMT) ("mov (%rax), %rax\n")		// Use a few cache misses as a delay.
		"mov %rax, delay_array_pointer\n"
#endif

// Use misaligned rmw to known addresses as a delay. Use this for Atom Silvermont, where loads are executed in-order (any loads with unknown address
// blocks all later loads). We want a high-latency event that doesn't block later loads, so use cache-missed loads that have
// known addresses. Must be careful that misaligned accesses aren't handled by a pipeflush.
#if 0
	REPH2(DELAYAMT) (
		"clflush delay_array_pointer\n"
		"addl $0, delay_array_pointer\n"
		"clflush delay_array_pointer\n"
		"addl $0, delay_array_pointer+1\n" )
#endif		
//		REPH2(DELAYAMT) ("imul %eax\n")
R"(		
		and $0x0, %rax		# I think this is needed only for paranoia in case uarch flushes pipeline early if ret completes early
		add %rax, (%rsp)
	delay_and_return_train:
		ret
		
		return_2:
			addq $2, (%rsp)
		return_2_train:
			ret
)");
#else
#error Unknown architecture
#endif

extern void cl0();
asm ( R"(
	.data
	.align 4096
	.skip 32	# Reduce cache way conflicts with 2-way 8KB Dcache on PPro
	.skip 8		# Reduce SNB/IVB L1 cache bank conflicts
	cl0: .skip 4096*6
	.text	
)" );


/*	Time a load of each cache line, to determine hit/miss.
	The timing routine is not entirely straightforward, to increase precision. 
	RDTSC/RDTSCP (on IVB, at least) has fairly high latency (~24ish clocks), so
	a small load latency (e.g., L1 or L2 hit) gets completely hidden and L1 vs. L2 hit can't
	be distinguished. Fences don't help: They're still high latency operations so the load
	latency is still completely hidden. This timing routine creates a data dependence
	chain from the rdtsc, through the load, then to the return address that causes a branch
	misprediction. The second rdtsc occurs immediately after the resolution of the mispredicted
	return. This makes the execution time of the second rdtsc directly tied to the load latency.
*/
extern unsigned int time_load(uintptr_t addr) {
	unsigned int d;
	asm volatile ( R"(
		call return_2_train
		call 4f
		call return_2_train
		call 4f				# Train the RAS (for Via Nano)
		call return_2; 1: jmp 1b		# Pipeline flush
		rdtsc				# Start timing
		and $4, %%edx		# Avoid using $0 just in case a future uarch breaks the data dependency. Existing uarchs don't seem to though.
	)"
#if __i386__					
		"mov (%[addr],%%edx), %%edx\n"	// Make load address data-dependent on rdtsc so it can't be reordered early
#elif __x86_64__
		"mov (%[addr],%%rdx), %%edx\n"
#endif
	R"(
		mov %%eax, %%esi
		call 1f
		2: jmp 2b			
		1: and $3f-2b, %%edx
		or $3f-2b, %%edx		# edx = 12
	)"
#if __i386__					
		"add %%edx, (%%esp)\n"	// Adjust return address to force a pipe flush, with branch misprediction resolution time dependent on the load latency
#elif __x86_64__
		"add %%rdx, (%%rsp)\n"
#endif
	R"(
		4: ret					# Returns to the rdtsc
		3: rdtsc				# End timing
		sub %%esi, %%eax
	)" : "=&a"(d) : [addr] "r"(addr) : "edx","esi");
	return d;
}

paddr_t cr3;
static unsigned int threshold = 0;

/*
	IVB: Instructions that don't work speculatively
	- clflush
	- prefetch
	- stores
	IVB: Speculative illegal load may return 0 on bank conflict (victim and target are at same offset within the cache line
*/
void spec_load(unsigned int *read_ptr, unsigned int counts[6][1024], int pin_method, unsigned int *cache_pin, int prefetch_only)
{
	memset (counts, 0, 6*1024*sizeof(int));
	unsigned int sum[6]={0,0,0,0,0,0};

#if (SPECLOAD_REPORT==SPECLOAD_VALUE)
	// List of values, sized for worst case
	unsigned int values[2][ITS] = {0};
	unsigned int values_last = 0;
#endif

	unsigned int cpuid_80000001_ecx=0;
#if !OLD_CPU
	asm volatile ("mov $0x80000001, %%eax; cpuid\n" : "=c"(cpuid_80000001_ecx) : : "eax","ebx","edx");		
#endif
	
	//paddr_t paddr = translate(cr3, (uintptr_t)read_ptr, 0);
	//printf ("Translated %p to paddr %016llx\n", read_ptr, paddr);
	
	
	for (int i=-20;i<ITS;i++) {
		// Prefetch victim by one of many methods
#if OLD_CPU
		pgmod_wbinvd();					// Sorry, dumping out the entire cache
#else		
#if (SPECLOAD_REPORT==SPECLOAD_VALUE)
	for (int j=0;j<20480;j+=32) {
		asm volatile ("clflush cl0(%0)\n" : : "r"(j));
	}
#elif (SPECLOAD_REPORT==SPECLOAD_STATUS)
		asm volatile ( R"(
			clflush cl0					# Flush cache lines
			clflush cl0+4096+128
			clflush cl0+8192+256
			clflush cl0+12288+384
			clflush cl0+16384+512
		)" );
#endif		
		if (cache_pin) {
			// We need to clflush the cache_pin address because we measure the load latency for it
			// later, which loads it into the L1 cache.
#if DONT_TRUST_CLFLUSH
			for (int k=0;k<4;k++) {
				asm volatile ( R"(
					mov -12288(%0), %%eax
					mov -4096(%0), %%eax
					mov -26624(%0), %%eax
					mov -24576(%0), %%eax
					mov -20480(%0), %%eax
					mov (%0), %%eax
					mov -6144(%0), %%eax
					mov -16384(%0), %%eax
					mov -14336(%0), %%eax
					mov -22528(%0), %%eax
					mov -2048(%0), %%eax
					mov -28672(%0), %%eax
					mov -18432(%0), %%eax
					mov -8192(%0), %%eax
					mov -10240(%0), %%eax
					mov -30720(%0), %%eax
				)" : : "r"(cache_pin) : "eax");
			}
#endif	// DONT_TRUST_CLFLUSH			
			asm volatile ( R"(
				clflush (%0)
			)" : : "r"(cache_pin));
		}

		asm volatile ( R"(
				mfence			
				call return_2_train
				call return_2_train
				call return_2; 1: jmp 1b		# Barrier
		)" );	
#endif
		sched_yield();		// Reduce context switches during the test.
		switch (pin_method) {
			case 0:	break;		// No prefetch, so it should be uncached.
			case 1: asm volatile ("mov (%0), %%eax\n" : : "r"(cache_pin) : "eax"); break;
			case 2: asm volatile ("mov %%eax, 8(%0)\n" : : "r"(cache_pin) : "eax"); break;
			case 3: asm volatile ("call delay_and_return_train; call delay_and_return_train; mov $3f-2f, %%eax ; call delay_and_return ; 2: mov (%0), %%eax; 1: jmp 1b; 3:\n" : : "r"(cache_pin) : "eax"); break;
			case 4: asm volatile ("call delay_and_return_train; call delay_and_return_train;  mov $3f-2f, %%eax ; call delay_and_return ; 2: mov %%eax, 8(%0); 1: jmp 1b; 3:\n" : : "r"(cache_pin) : "eax"); break;
			case 10: asm volatile ("call delay_and_return_train; call delay_and_return_train;  mov $3f-2f, %%eax ; call delay_and_return ; 2: mov (%0), %%eax; 1: jmp 1b; 3:\n" : : "r"(read_ptr) : "eax"); break;
			case 11: asm volatile ("call delay_and_return_train; call delay_and_return_train;  mov $3f-2f, %%eax ; call delay_and_return ; 2: mov %%eax, 8(%0); 1: jmp 1b; 3:\n" : : "r"(read_ptr) : "eax"); break;
			// Use 18 and 19 only if read_ptr is accessible.
			case 18: asm volatile ("mov (%0), %%eax\n" : : "r"(read_ptr) : "eax"); break;
			case 19: asm volatile ("mov %%eax, 8(%0)\n" : : "r"(read_ptr) : "eax"); break;
			case 20: pgmod_kernel_load(cache_pin); break;
			case 21: pgmod_kernel_load(read_ptr); break;
#if !OLD_CPU
			case 5: asm volatile ("prefetcht0 (%0)\n" : : "r"(cache_pin) : "eax"); break;
			case 6: asm volatile ("prefetcht1 (%0)\n" : : "r"(cache_pin) : "eax"); break;
			case 7: asm volatile ("prefetcht2 (%0)\n" : : "r"(cache_pin) : "eax"); break;
			case 8: asm volatile ("prefetchnta (%0)\n" : : "r"(cache_pin) : "eax"); break;
			case 9: if (cpuid_80000001_ecx & (1<<8)) { asm volatile ("prefetchw (%0)\n" : : "r"(cache_pin) : "eax"); }
				else { printf ("No support for prefetchw\n"); return; }
				break;
			case 12: asm volatile ("prefetcht0 (%0)\n" : : "r"(read_ptr) : "eax"); break;
			case 13: asm volatile ("prefetcht1 (%0)\n" : : "r"(read_ptr) : "eax"); break;
			case 14: asm volatile ("prefetcht2 (%0)\n" : : "r"(read_ptr) : "eax"); break;
			case 15: asm volatile ("prefetchnta (%0)\n" : : "r"(read_ptr) : "eax"); break;
			case 16: if (cpuid_80000001_ecx & (1<<8)) { asm volatile ("prefetchw (%0)\n" : : "r"(read_ptr) : "eax"); }
				else { printf ("No support for prefetchw\n"); return; }
				break;
			case 17: asm volatile ("mov (%0), %%eax; mfence; prefetcht0 (%1)\n" : : "r"(cache_pin),"r"(read_ptr) : "eax"); break;
			
			case 100: asm volatile ("clflush (%0)\n" : : "r"(cache_pin) : "eax"); break;
#endif			
			default: printf ("Prefetch method %d not supported\n", pin_method); return; break;
		}

		// Give the prefetch a lot of time to complete. mfence alone doesn't seem to fence prefetches.
		excessive_fence();
		if (!prefetch_only) {
			asm volatile ( R"(
				call delay_and_return_train
				call delay_and_return_train
				mov $nonspec_return_target-2f, %%eax	# Set return address
				
				call delay_and_return
					# Speculative execution
					2: 
					mov (%0), %%eax			# Read victim
				)"
#if (SPECLOAD_REPORT==SPECLOAD_VALUE)		// Try to report the 32-bit value that's returned
				R"(
					mov %%eax, %%ecx
					shl $6, %%ecx
					and $0x7c0, %%ecx
					mov cl0+0(%%ecx), %%ecx
					
					mov %%eax, %%ecx
					shl $1, %%ecx
					and $0x7c0, %%ecx
					mov cl0+2048(%%ecx), %%ecx

					mov %%eax, %%ecx
					shr $4, %%ecx
					and $0x7c0, %%ecx
					mov cl0+4096(%%ecx), %%ecx

					mov %%eax, %%ecx
					shr $9, %%ecx
					and $0x7c0, %%ecx
					mov cl0+6144(%%ecx), %%ecx

					mov %%eax, %%ecx
					shr $14, %%ecx
					and $0x7c0, %%ecx
					mov cl0+8192(%%ecx), %%ecx
					
					mov %%eax, %%ecx
					shr $19, %%ecx
					and $0x7c0, %%ecx
					mov cl0+10240(%%ecx), %%ecx
					
					mov %%eax, %%ecx
					shr $24, %%ecx
					and $0x7c0, %%ecx
					mov cl0+12288(%%ecx), %%ecx					
					1: jmp 1b
				)"
#elif (SPECLOAD_REPORT==SPECLOAD_STATUS)		// Report whether the value is 0 or 4224, with extra paranoia to test for stalls
#if VIA_NANO				
					"add $0, %%eax\n"		// Produces different result on Via Nano. Load->load causes the second load to be not executed, but will execute with value 0 if Load->ALU instead.
#endif
				R"(
					mov cl0(%%eax), %%ecx	# Determine value of victim
					and $4, %%eax	# eax is now 0 or 4
					mov cl0+12288+384(%%eax), %%ecx		# 3: Detect whether a value was returned
					mov cl0+16384+512, %%ecx			# 4: Detect whether this code even executed					
					1: jmp 1b
				)"
#else
#error Invalid value for SPECLOAD_REPORT. Must be SPECLOAD_STATUS or SPECLOAD_VALUE
#endif				
				"nonspec_return_target:\n"
			: : "r"(read_ptr) : "eax","ecx");
		}

		for (int i=0;i<10000;i++) asm volatile ("nop");

		//time_load(((uintptr_t)cl0) + 16384+896);		// First execution is slower by ~5 clocks, so do a dummy timing on an unrelated location first.
		if (i >= 0) {
#if (SPECLOAD_REPORT==SPECLOAD_VALUE)
			unsigned int hits[7] = {0};		// 7 groups of 5 bits = 35 bits.
			unsigned int hitval[7] = {0};
			for (int k=0; k < 7; k++) {
				for (int j=0;j<32;j++) {
					unsigned int d;				
					uintptr_t addr;
					unsigned int mix = (j*91)&0x1f;
					addr = ((uintptr_t)cl0) + (k*2048) + (mix<<6);
					d = time_load(addr);
					if (d < threshold) {
						hits[k]++;
						hitval[k] = mix << (k*5);
					}
				}
			}
			int ok = 1;
			unsigned int val = 0;
			for (int j=0;j<7;j++) {
				if (hits[j] != 1) { ok = 0; break; }
				val |= hitval[j];
			}
			// printf ("Hits: %u %u %u %u %u %u %u\n", hits[0], hits[1], hits[2], hits[3], hits[4], hits[5], hits[6]);
			if (ok) {
				int j;
				for (j=0;j<values_last;j++) {				
					if (values[0][j] == val)
						break;
				}
				values[0][j] = val;
				values[1][j]++;
				if (j == values_last) {
					values_last++;
				}
			}
#elif (SPECLOAD_REPORT==SPECLOAD_STATUS)
#else
#error Invalid value for SPECLOAD_REPORT. Must be SPECLOAD_STATUS or SPECLOAD_VALUE		
#endif			
			for (int j=0;j<6;j++) {
				unsigned int d;				
				uintptr_t addr;
				if (j < 5) addr = ((uintptr_t)cl0) + j*4224;
				else if (j == 5) addr = (uintptr_t)cache_pin;
				d = time_load(addr);				
				sum[j] += d;
				if (d < 1023)
					counts[j][d]++;
				else
					counts[j][1023]++;
			}
		}
	}
#if (SPECLOAD_REPORT==SPECLOAD_VALUE)
	for (int i=0;i<values_last;i++) {
		printf ("\t%d: %08x\n", values[1][i], values[0][i]);
	}
	printf ("\n");
#elif (SPECLOAD_REPORT==SPECLOAD_STATUS)
	printf ("Avg:\t%.1f\t%.1f\t%.1f\t%.1f\t%.1f\t%.1f\t", (float)sum[0]/ITS, (float)sum[1]/ITS, (float)sum[2]/ITS, (float)sum[3]/ITS, (float)sum[4]/ITS, (float)sum[5]/ITS);
	unsigned int min[6]={0,0,0,0,0,0}, max[6]={0,0,0,0,0,0}, sum2[6]={0,0,0,0,0,0};
	for (int i=0;i<1024;i++) {
		for (int j=0;j<6;j++) {
			sum2[j] += counts[j][i];
			if (min[j] == 0 && sum2[j] >= 0.02 * ITS) min[j] = i;
			if (max[j] == 0 && sum2[j] >= 0.98 * ITS) max[j] = i;
		}
	}
	//printf ("\n 2%%:\t%u\t%u\t%u\t%u\t%u\t%u\n", min[0], min[1], min[2], min[3], min[4], min[5]);
	//printf ("98%%:\t%u\t%u\t%u\t%u\t%u\t%u\t", max[0], max[1], max[2], max[3], max[4], max[5]);
	if (!prefetch_only) {
		int t = ((min[0] < threshold) ? 1 : 0) + ((min[1] < threshold) ? 2 : 0); 
		printf ("\t%s\n", 
			(min[4] >= threshold) ? "Not executed (or in-onder stall)" : 
			(t==0) ? ((min[3] < threshold) ? "Unknown value" : "Stall") : 
			(t==1) ? "Zero" :
			(t==2) ? "Read" :
			(t==3) ? "Error1" : "Error2" );
	}
	else {
		printf ("\n");
	}
#else
#error Invalid value for SPECLOAD_REPORT. Must be SPECLOAD_STATUS or SPECLOAD_VALUE
#endif
}


void do_test(unsigned int *two_values, int pin_method, unsigned int *cache_pin, int prefetch_only)
{
	unsigned int counts[6][1024];

	spec_load(two_values, counts, pin_method, cache_pin, prefetch_only);
	
	unsigned int sum0=0, sum1=0, maxdiff=0, threshold=0;
	float error_rate0, error_rate1;
	for (int i=0;i<1024;i++) {
		sum0 += counts[0][i];
		sum1 += counts[1][i];
		unsigned int d = abs(sum0 - sum1);
		if (d > maxdiff) {
			maxdiff = d;
			threshold = i;
			error_rate0 = (float)sum0 / ITS;
			error_rate1 = (float)sum1 / ITS;
			if (sum0 > sum1)
				error_rate0 = 1.0 - error_rate0;
			else
				error_rate1 = 1.0 - error_rate1;
		}
	}
	//printf ("Threshold %u, errror rates = %.6f %.6f (avg %.4f)\n", threshold, error_rate0, error_rate1, (error_rate0 + error_rate1) / 2);
}

unsigned int find_threshold (unsigned int *two_values, unsigned int *cache_pin)
{
	unsigned int counts[6][1024];

	spec_load(two_values, counts, 1, cache_pin, 1);
	unsigned int sum0=0, sum1=0;
	for (int i=0;i<1024;i++) {
		sum0 += i*counts[4][i];
		sum1 += i*counts[5][i];
	}
	unsigned int threshold = sqrt( ((float)sum0)*((float)sum1) ) / ITS;
	printf ("Cache hit vs. miss threshold: %u\n", threshold);
	return threshold;
}

static inline unsigned long long my_rand (unsigned long long limit)
{
	if (limit == 0) return 0;
	return ((unsigned long long)(((unsigned long long)rand()<<48)^((unsigned long long)rand()<<32)^((unsigned long long)rand()<<16)^(unsigned long long)rand())) % limit;
}
void init_delay_array()
{
	const unsigned int size = 67108864;
	delay_array = mmap(0, size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
	if (delay_array==MAP_FAILED) {
		perror ("Failed to mmap 64 MB\n");
		exit(1);
	}
	for (unsigned int i=0; i < size/4096; i++) {
		int idx = (my_rand(i)*4096 + DELAY_OFFSET)/sizeof(void*);
		delay_array[(i*4096+DELAY_OFFSET)/sizeof(void*)] = delay_array[idx];
		delay_array[idx] = &delay_array[(i*4096+DELAY_OFFSET)/sizeof(void*)];
	}
	delay_array_pointer = &delay_array[DELAY_OFFSET/sizeof(void*)];
}

jmp_buf jmpbuf;
void quitnow(int signo) {
	printf ("Quitting now %d\n", signo);
	longjmp (jmpbuf, 1);
}


int main(int argc, char* argv[])
{
	init_delay_array();

	printf ("ALIAS_DISTANCE: %u\n", ALIAS_DISTANCE);
	printf ("DONT_REMAP_PTE1: %u\n", DONT_REMAP_PTE1);
	printf ("DONT_TRUST_CLFLUSH: %u\n", DONT_TRUST_CLFLUSH);
	printf ("DELAYAMT: %u\n", DELAYAMT);

	/* Use pgmod to set up the test
	   /proc/pgmod - Used to retrieve CR3 and CR4, and reload CR3 to cause a TLB flush
	   /dev/pgmod_mem - Same as /dev/mem but bypasses all security restrictions (e.g., CONFIG_STRICT_DEVMEM)
	   
	   This is used to allow control over the page tables. We allocate two memory pages (using mmap), then
	   change the page tables for that array so both virtual pages point to the same physical page. This is the
	   "victim" page that contains data we want to read. The first virtual page's permissions are changed to see
	   how permissions affects the ability to read the victim. The second virtual page's permissions are
	   not restricted, and is used to preload/pin the physical page into various (mainly L1) caches.
	   
	   On Intel CPUs, the ability to read from the victim page where permissions checks fail requires that the
	   data actually be in the L1 cache. The second page is used to ensure this is the case.
	*/
	pgutil_info_struct pginfo;
	if (pgutil_init(&pginfo) != 0) {
		return -1;
	}
	paddr_t cr3 = pginfo.cr3;
	uint64_t patmsr = pginfo.patmsr;
	
	char pat_table[8][4] = {
		"UC", "WC", "--", "--", "WT", "WP", "WB", "UC-"
	};
	for (int p=0;p<8;p++) {
		printf ("  %d: %s\n", p, pat_table[ (patmsr>>(p*8)) & 0x7 ]);
	}
	
	
	unsigned int* array = (unsigned int*)mmap(0, ALIAS_DISTANCE*4+8192, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
	if (array==MAP_FAILED) {
		printf ("Failed to mmap %u KB\n", ALIAS_DISTANCE/256+8);
		exit(1);
	}
	madvise(array, ALIAS_DISTANCE*4+8192, MADV_NOHUGEPAGE);		// Transparent hugepage collapsing can fail because we changed the page tables?

	array[0] = 4096+128;		// First write to page causes a copy, so don't do translation until after writing. (Copy on write?)
	array[1] = 8192+256;
	array[ALIAS_DISTANCE]=8192+256;		// Make this value different from array[0] so we can test whether page table remapping worked.
								// Place the alias 64KB away. On K7 (64 KB 2 way L1), we want the alias at a multiple of 32 KB
								// away so both aliases hit in the L1 instead of evicting each other out of the L1 (VIPT cache with alias detection)
	memset(cl0, 0, 6*4096);		// Write over cl0 pages, in case there's some funny demand-paging stuff happening (doesn't seem necessary).

	// Sanity check:
	paddr_t pv = translate(cr3, (uintptr_t)array, 0);
	unsigned int pvalue;
	read_physical(pv, &pvalue, 4);
	printf ("Sanity check: Victim page is at virtual address %p, physical address %016llx, with value = %d\n", array, (uint64_t)pv, pvalue);

	// Change the PTEs for the two pages
	const pte_t orig_pte0 = translate(cr3, (uintptr_t)array, 1);
	const pte_t orig_pte1 = translate(cr3, ((uintptr_t)array)+ALIAS_DISTANCE*4, 1);
	const paddr_t pte0_addr = translate(cr3, (uintptr_t)array, -1);
	const paddr_t pte1_addr = translate(cr3, ((uintptr_t)array)+ALIAS_DISTANCE*4, -1);
	printf ("Original PTE0 is %016llx\n", (uint64_t)orig_pte0);
	printf ("Original PTE1 is %016llx\n", (uint64_t)orig_pte1);
	
	if (!setjmp(jmpbuf)) {
	
		signal(SIGINT, quitnow);
		signal(SIGSEGV, quitnow);
	
		pte_t new_pte0 = orig_pte0;
		write_physical(pte0_addr, &new_pte0, sizeof(pte_t));		// Change PTE
		if (!DONT_REMAP_PTE1) write_physical(pte1_addr, &orig_pte0, sizeof(pte_t));
		pgmod_flush_tlb();
		
		printf ("Checking that victim is accessible through the alias (expected yes): %s (%d == %d)\n", pvalue == array[ALIAS_DISTANCE] ? "yes" : "no", pvalue, array[ALIAS_DISTANCE]);
		printf ("The two virtual addresses are %p and %p, spaced apart by %08lx\n", &array[0], &array[ALIAS_DISTANCE], (uintptr_t)&array[ALIAS_DISTANCE] - (uintptr_t)&array[0]);

		threshold = find_threshold(&array[0], &array[ALIAS_DISTANCE]);


		// Change permissions of the page containing "value", then see whether we can still read it.
		// write and present bits
		// cacheability bits (memory type)	
		// Whether value is in the data caches, and DTLB miss.
		// A, D, G bits?
		// Prefetch methods
	#if 1
		printf ("Prefetch methods (Accessible):\n");
		for (int method = 0; method <= 21; method++) {
			printf ("%u: ", method);
			do_test(&array[0], method, &array[ALIAS_DISTANCE], 1);
		}

		new_pte0 &= ~0x4ULL;	// Supervisor only
		write_physical(pte0_addr, &new_pte0, sizeof(pte_t));		// Change PTE
		pgmod_flush_tlb();
		printf ("Prefetch methods (Supervisor only):\n");
		for (int method = 0; method <= 21; method++) {
			if (method == 18 || method == 19) continue;
			printf ("%u: ", method);
			do_test(&array[0], method, &array[ALIAS_DISTANCE], 1);
		}
	#endif
	
		unsigned int ptebits[] = {
			0x67,	// Accessible
			0x66,	// Not present
			0x00,	// Really not present
	//		0x65,	// Not writable
			0x63,	// Supervisor only
			0x47,	// Not accessed
	//		0x27,	// Not dirty
			0x43,	// Not accessed *and* supervisor only
			0x6f,	// Accessible, write-through (WC?)
			0x77,	// Accessible, cache-disable (UC-)
			0x7f,	// Accessible, CD+WT (UC)
			0xe7,	// Accessible, PAT (WT on Kernel 4.2+). 4KB pages only. 2M pages uses bit 12 as the PAT bit.
			0xef,	// Accessible, PAT+WT (what type is this?)
			0xf7,	// Accessible, PAT+WT (what type is this?)
			0xff,	// Accessible, PAT+WT (what type is this?)

			0x6b,	// Supervisor, write-through (WC?)
			0x73,	// Supervisor, cache-disable (UC-)
			0x7b,	// Supervisor, CD+WT (UC)
			0xe3,	// Supervisor, PAT (WT on Kernel 4.2+). 4KB pages only. 2M pages uses bit 12 as the PAT bit.
			0xeb,	// Supervisor, PAT+WT (what type is this?)
			0xf3,	// Supervisor, PAT+WT (what type is this?)
			0xfb	// Supervisor, PAT+WT (what type is this?)
		};

		unsigned int ptebits1[] = {
			//0x67,	// Accessible
			0x63,	// Supervisor only
			0x43,	// Not accessed *and* supervisor only
		};
	
		for (int mode = 0; mode < sizeof(ptebits)/sizeof(ptebits[0]) ; mode++) {		
			printf ("PTE bits %02x (%s%s%s%s%s%s):\n", ptebits[mode], pat_table[(patmsr >> (8* (((ptebits[mode]>>3)&3) | ((ptebits[mode]>>5)&4)))) & 7],
				(ptebits[mode]&4) ? ", User" : ", Supervisor",
				(ptebits[mode]&1) ? "" : ", Not present",
				(ptebits[mode]&2) ? "" : ", Not writable",
				(ptebits[mode]&0x20) ? "" : ", A=0",
				(ptebits[mode]&0x40) ? "" : ", D=0"			
			 );
			if ((ptebits[mode]&0x80) && !pginfo.pat) {
				printf ("  PAT not supported, skipping mode with PAT bit = 1\n\n");
			}
			else {		 
				pte_t pte = (new_pte0 & ~0xffULL) | ptebits[mode];
				if (ptebits[mode] == 0x0)
					pte = 0;		// Really try to clobber the PTE
				write_physical(pte0_addr, &pte, sizeof(pte_t));		// Change PTE
				pgmod_flush_tlb();						if (!(ptebits[mode]&1))
					printf("Don't kernel-touch a non-present page\n");
				else if (!(ptebits[mode]&0x20))
					printf("Don't kernel-touch a non-accessed page because it will become accessed and the result becomes meaningless.\n");
				else
					do_test(&array[0], 21, &array[ALIAS_DISTANCE], 0);		// L1 cache
					
				do_test(&array[0], 1, &array[ALIAS_DISTANCE], 0);		// L1 cache
				do_test(&array[0], 6, &array[ALIAS_DISTANCE], 0);		// L2 cache
				do_test(&array[0], 0, &array[ALIAS_DISTANCE], 0);		// Not in cache
			}
		}
		{
			unsigned int mode = 0x67;
			printf ("PTE bits %02x (%s%s%s%s%s%s), Using test_ptr for both setup and test:\n", mode, pat_table[(patmsr >> (8* (((mode>>3)&3) | ((mode>>5)&4)))) & 7],
				(mode&4) ? ", User" : ", Supervisor",
				(mode&1) ? "" : ", Not present",
				(mode&2) ? "" : ", Not writable",
				(mode&0x20) ? "" : ", A=0",
				(mode&0x40) ? "" : ", D=0"			
			 );
			new_pte0 = (new_pte0 & ~0xffULL) | mode;
			write_physical(pte0_addr, &new_pte0, sizeof(pte_t));		// Change PTE
			pgmod_flush_tlb();		
			do_test(&array[0], 21, &array[0], 0);		// L1 cache
			do_test(&array[0], 1, &array[0], 0);		// L1 cache
			do_test(&array[0], 6, &array[0], 0);		// L2 cache
			do_test(&array[0], 0, &array[0], 0);		// Not in cache
		}
		{
			unsigned int mode = 0x67;
			printf ("PTE bits %02x (%s%s%s%s%s%s), Using setup_ptr for both setup and test:\n", mode, pat_table[(patmsr >> (8* (((mode>>3)&3) | ((mode>>5)&4)))) & 7],
				(mode&4) ? ", User" : ", Supervisor",
				(mode&1) ? "" : ", Not present",
				(mode&2) ? "" : ", Not writable",
				(mode&0x20) ? "" : ", A=0",
				(mode&0x40) ? "" : ", D=0"			
			 );
			new_pte0 = (new_pte0 & ~0xffULL) | mode;
			write_physical(pte0_addr, &new_pte0, sizeof(pte_t));		// Change PTE
			pgmod_flush_tlb();		
			do_test(&array[ALIAS_DISTANCE], 21, &array[ALIAS_DISTANCE], 0);		// L1 cache
			do_test(&array[ALIAS_DISTANCE], 1, &array[ALIAS_DISTANCE], 0);		// L1 cache
			do_test(&array[ALIAS_DISTANCE], 6, &array[ALIAS_DISTANCE], 0);		// L2 cache
			do_test(&array[ALIAS_DISTANCE], 0, &array[ALIAS_DISTANCE], 0);		// Not in cache
		}		
		
	}



	// Restore PTE before quitting, or Linux will complain.
	write_physical(pte0_addr, &orig_pte0, sizeof(pte_t));
	write_physical(pte1_addr, &orig_pte1, sizeof(pte_t));
	
	
	pgmod_flush_tlb();
	printf ("Restored PTEs\n");

	printf ("Checking that victim is accessible through the alias (expected no): %s (%d == %d)\n", pvalue == array[ALIAS_DISTANCE] ? "yes" : "no", pvalue, array[ALIAS_DISTANCE]);

	munmap(array, ALIAS_DISTANCE*4+8192);

	pgutil_cleanup();	
	return 0;
}



